/** TileMap2D.cpp
 *
 * See the header file for details.
 */

#include "TileMap2D.h"

#include "TileMap2DLoader.h"

using namespace raylib;

static raylib::Rectangle toRayLibRect(tson::Rect rect) {
	return raylib::Rectangle(rect.x, rect.y, rect.width, rect.height);
}

TileMap2D::TileMap2D(const std::string &fileName) {
	tileMap = loadTileMap(fileName);
	
	if(tileMap->getStatus() != tson::ParseStatus::OK) {
		throw std::runtime_error(TextFormat("Failed to load tile-map %s. Error: %s", 
			fileName.c_str(), tileMap->getStatusMessage().c_str()));
	}
	
	// Load the images
	std::string baseDir = GetDirectoryPath(fileName);
	auto layers = tileMap->getLayers();
	for (const auto& layer : layers) {
        if (layer.getType() == tson::LayerType::ImageLayer) {
            loadImage(baseDir, layer.getImage());
        }
    }
	auto tilesets = tileMap->getTilesets();
	for (auto& tileset : tilesets) {
		auto &image = tileset.getImage();
		if(!image.empty()) {
			loadImage(baseDir, image.string());
		} else {
			for(const auto& tile : tileset.getTiles()) {
				auto &image = tile.getImage();
				if(!image.empty()) {
					loadImage(baseDir, image.string());
				}
			}
		}
    }
}

TileMap2D::~TileMap2D() {
}

void TileMap2D::draw(const raylib::Camera2D &camera, const raylib::Color &tint) {
	auto layers = tileMap->getLayers();
	for(auto &layer : layers) {
		drawLayer(layer, camera, raylib::Vector2(0,0), tint);
	}
}

void TileMap2D::updateAnimations(float elapsedTime) {
	// ##### FIXME! ###### To-do
}

raylib::Vector2 TileMap2D::getMapSize() const {
	auto size = tileMap->getSize();
	auto tileSize = tileMap->getTileSize();
	return raylib::Vector2(size.x * tileSize.x, size.y * tileSize.y);
}

void TileMap2D::drawLayer(
		tson::Layer &layer, const raylib::Camera2D &camera, const raylib::Vector2 &offset, const raylib::Color &tint) {
	auto pFactorTson = layer.getParallax();
	auto pOriginTson = layer.getMap()->getParallaxOrigin();
	raylib::Vector2 parallaxFactor(pFactorTson.x, pFactorTson.y);
	raylib::Vector2 parallaxOrigin(pOriginTson.x, pOriginTson.y);
	
	auto layerOffsetTson = layer.getOffset();
	raylib::Vector2 layerOffset(layerOffsetTson.x, layerOffsetTson.y);
	auto cameraOffset = camera.GetWorldToScreen(parallaxOrigin) + raylib::Vector2(-GetScreenWidth() / 2, -GetScreenHeight() / 2);
	
	raylib::Vector2 newOffset = offset + layerOffset -
		 cameraOffset * (raylib::Vector2(1.0f, 1.0f) - parallaxFactor);
	
	switch(layer.getType()) {
	case tson::LayerType::TileLayer:
		drawTileLayer(layer, camera, newOffset, tint);
		break;
	case tson::LayerType::ImageLayer:
		drawImageLayer(layer, camera, newOffset, tint);
		break;
	case tson::LayerType::Group:
	{
		for(auto &subLayer : layer.getLayers()) {
			drawLayer(subLayer, camera, newOffset, tint);
		}
		break;
	}
	default:
		// Fail
		auto *errMsg = TextFormat("ERROR: Unsupported Tileson layer type %u", 
			layer.getType());
		throw std::runtime_error(errMsg);
	}
}
	
void TileMap2D::drawTileLayer(
		tson::Layer &layer, const raylib::Camera2D &camera, const raylib::Vector2 &offset, const raylib::Color &tint) {
	for (auto& [pos, tileObject] : layer.getTileObjects()) {
		tson::Tile* tile = tileObject.getTile();
		tson::Tileset *tileset = tile->getTileset();
		auto tileImage = tileset->getImage();
		auto imagePath = tileImage.empty() ? tile->getImage() : tileImage;
		auto texIter = textures.find(imagePath.string());
		if(texIter == textures.end()) {
			auto *errMsg = TextFormat("ERROR: Tile-map is missing image: %s\n", 
				imagePath.string().c_str());
			throw std::runtime_error(errMsg);
		}
		auto &texture = texIter->second;
		
		raylib::Rectangle drawRect = toRayLibRect(tileObject.getDrawingRect());
		auto tilePos = tileObject.getPosition();
		raylib::Vector2 drawPos(tilePos.x + offset.x, tilePos.y + offset.y);
		texture->Draw(drawRect, drawPos, tint);
	}
}
	
void TileMap2D::drawImageLayer(
		tson::Layer &layer, const raylib::Camera2D &camera, const raylib::Vector2 &offset, const raylib::Color &tint) {
	auto imageName = layer.getImage();
	auto texIter = textures.find(imageName);
	if(texIter == textures.end()) {
		auto *errMsg = TextFormat("ERROR: Tile-map is missing image: %s\n", 
			imageName.c_str());
		throw std::runtime_error(errMsg);
	}
	auto &texture = texIter->second;
	
	bool repeatX = layer.hasRepeatX();
	bool repeatY = layer.hasRepeatY();
	
	if(repeatX || repeatY) {
		auto screenOriginPos = 
			camera.GetScreenToWorld(raylib::Vector2(0.0f,0.0f));
		auto texWidth = texture->GetWidth();
		auto texHeight = texture->GetHeight();
		raylib::Rectangle sourceRec(
			0, 0, texWidth, texHeight); 
		raylib::Rectangle destRec(
			repeatX ? screenOriginPos.x : offset.x, 
			repeatY ? screenOriginPos.y : offset.y,
			repeatX ? GetScreenWidth() : texWidth, 
			repeatY ? GetScreenHeight() : texHeight);
		raylib::Vector2 origin = raylib::Vector2(0,0);
		texture->DrawTiled(sourceRec, destRec, origin);
	} else {
		texture->Draw(offset, tint);
	}
}

void TileMap2D::loadImage(const std::string &baseDir, const std::string &fileName) {
	if(textures.find(fileName) != textures.end()) {
		// Already loaded
		return;
	}
	
	std::string fullPath = (baseDir.length() > 0) ?
		TextFormat("%s/%s", baseDir.c_str(), fileName.c_str()) :
		fileName;
	
	textures[fileName] = std::make_shared<raylib::Texture>(fullPath);
}

