/** TrackingCamera2D.cpp
 *
 * See header file for details.
 */

#include "TrackingCamera2D.h"

#define DEFAULT_P_FACTOR 0.005f
#define DEFAULT_I_FACTOR 0.0025f
#define DEFAULT_TARGET_RADIUS 100.0f

TrackingCamera2D::TrackingCamera2D() {
	init();
}

TrackingCamera2D::TrackingCamera2D(::Vector2 offset, ::Vector2 target,
            float rotation, float zoom) : 
			raylib::Camera2D(offset, target, rotation, zoom) {
	position = target;
	init();
}

void TrackingCamera2D::setTarget(std::shared_ptr<Actor> target) {
	this->target = target;
}
	
std::shared_ptr<Actor> TrackingCamera2D::getTarget() {
	return target;
}

void TrackingCamera2D::setSceneSize(const raylib::Vector2 &sceneSize) {
	this->sceneSize = sceneSize;
}

bool TrackingCamera2D::update(bool onGround) {
	if(!target) {
		return true;
	}
	
	// Using the photography "rule of thirds" horizontally,
	// so calculate the offset based on the direction that the
	// target is moving
	if(target->velocity.x > 0) {
		int screenWidth = GetScreenWidth();
		targetOffset = -raylib::Vector2(screenWidth / 6.0f, 0.0f);
	} else if(target->velocity.x < 0) {
		int screenWidth = GetScreenWidth();
		targetOffset = raylib::Vector2(screenWidth / 6.0f, 0.0f);
	}

	// This is a basic PI controller. See:
	// https://en.wikipedia.org/wiki/PID_controller
	auto error = target->position - (this->position + targetOffset);
	auto targetVelocity = target->velocity;
	
	if(sceneSize.Length() > 0) {
		// Clamp the camera to the scene size
		auto cameraOffset = GetOffset();
		if(position.x - cameraOffset.x <= 0) { 
			position.x = cameraOffset.x; 
			if(error.x < 0) { 
				error.x = 0.0f; 
				integral.x = 0.0f;
				targetVelocity.x = 0.0f;
			}
		}
		if(position.y - cameraOffset.y <= 0) { 
			position.y = cameraOffset.y; 
			if(error.y < 0) { 
				error.y = 0.0f; 
				integral.y = 0.0f;
				targetVelocity.y = 0.0f;
			}
		}
		if(position.x + cameraOffset.x >= sceneSize.x) { 
			position.x = sceneSize.x - cameraOffset.x; 
			if(error.x > 0) { 
				error.x = 0.0f; 
				integral.x = 0.0f;
				targetVelocity.x = 0.0f;
			}
		}
		if(position.y + cameraOffset.y >= sceneSize.y) { 
			position.y = sceneSize.y - cameraOffset.y;
			if(error.y > 0) { 
				error.y = 0.0f; 
				integral.y = 0.0f;
				targetVelocity.y = 0.0f;
			}
		}
	}
	
	if(error.Length() < targetRadius) {
		integral = raylib::Vector2::Zero();
	}
	integral += error;
	velocity = error * propFactor + integral * integralFactor
		+ targetVelocity;
	
	this->position += this->velocity;
	
	SetTarget(this->position);
	
	return Actor::update(onGround);
}

void TrackingCamera2D::init() {
	sceneSize = raylib::Vector2(0.0f, 0.0f);
	propFactor = DEFAULT_P_FACTOR;
	integralFactor = DEFAULT_I_FACTOR;
	targetRadius = DEFAULT_TARGET_RADIUS;
	integral = raylib::Vector2(0.0f, 0.0f);
	
	int screenWidth = GetScreenWidth();
	targetOffset = -raylib::Vector2(screenWidth / 6.0f, 0.0f);
	setAffectedByGravity(false);
}

